/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Color.hh"

using namespace SharpConstruct;

Color::Color()
{
	Set( 1, 1, 1, 0 );
}

Color::Color( const Color& c )
{
	Set( c.R(), c.G(), c.B(), c.Mask() );
}

Color::Color( const float r, const float g, const float b )
{
	Set( r, g, b, 0 );
}

Color::Color( const float r, const float g, const float b, const float m )
{
	Set( r, g, b, m );
}

Color::Color( const Gdk::Color& c )
{
	Set( c.get_red_p(),
	     c.get_green_p(),
	     c.get_blue_p(), 1 );
}

Gdk::Color Color::GdkColor() const
{
	Gdk::Color col;
	col.set_rgb_p( R(), G(), B() );
	return col;
}

float Color::R() const
{
	return components_[ 0 ];
}
float Color::G() const
{
	return components_[ 1 ];
}
float Color::B() const
{
	return components_[ 2 ];
}
float Color::Mask() const
{
	return components_[ 3 ];
}

const float* Color::Data() const
{
	return components_;
}
float* Color::Data()
{
	return components_;
}

void Color::Set( const float r, const float g, const float b, const float m )
{
	SetRGB( r, g, b );
	SetMask( m );
}
void Color::SetRGB( const float r, const float g, const float b )
{
	SetR( r );
	SetG( g );
	SetB( b );
}
void Color::SetR( const float r )
{
	components_[ 0 ] = r;
}
void Color::SetG( const float g )
{
	components_[ 1 ] = g;
}
void Color::SetB( const float b )
{
	components_[ 2 ] = b;
}
void Color::SetMask( const float m )
{
	components_[ 3 ] = m;
	Clamp( components_[ 3 ] );
}

void Color::Clamp( float& c ) const
{
	if( c < 0.0f )
		c = 0.0f;
	else if( c > 1.0f )
		c = 1.0f;
}

void Color::operator=( const Color& c )
{
	Set( c.R(), c.G(), c.B(), c.Mask() );
}

Color Color::operator+( const Color& c ) const
{
	return Color( R() + c.R(), G() + c.G(), B() + c.B() );
}
Color Color::operator+( const float c ) const
{
	return Color( R() + c, G() + c, B() + c );
}
Color Color::operator-( const Color& c ) const
{
	return Color( R() - c.R(), G() - c.G(), B() - c.B() );
}
Color Color::operator-( const float c ) const
{
	return Color( R() - c, G() - c, B() - c );
}
Color Color::operator*( const float c ) const
{
	return Color( R() * c, G() * c, B() * c );
}

void Color::operator+=( const Color& c )
{
	SetR( R() + c.R() );
	SetG( G() + c.G() );
	SetB( B() + c.B() );
}
void Color::operator/=( const int d )
{
	SetR( R() / d );
	SetG( G() / d );
	SetB( B() / d );
}

Color SharpConstruct::MixColors( const Color& c1, const Color& c2 )
{
	return Color( ( c1.R() + c2.R() ) / 2,
		      ( c1.G() + c2.G() ) / 2,
		      ( c1.B() + c2.B() ) / 2,
		      ( c1.Mask() + c2.Mask() ) / 2 );
}
Color SharpConstruct::MixColors( const Color& c1, const Color& c2,
                                 const Color& c3 )
{
	return Color( ( c1.R() + c2.R() + c3.R() ) / 3,
		      ( c1.G() + c2.G() + c3.G() ) / 3,
		      ( c1.B() + c2.B() + c3.B() ) / 3,
		      ( c1.Mask() + c2.Mask() + c3.Mask() ) / 3 );
}
Color SharpConstruct::MixColors( const Color& c1, const Color& c2,
                                 const Color& c3, const Color& c4 )
{
	return Color( ( c1.R() + c2.R() + c3.R() + c4.R() ) / 4,
		      ( c1.G() + c2.G() + c3.G() + c4.G() ) / 4,
		      ( c1.B() + c2.B() + c3.B() + c4.B() ) / 4,
		      ( c1.Mask() + c2.Mask() + c3.Mask() + c4.Mask() ) / 4 );
}
