/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Brush.hh"
#include "Create.hh"
#include "DataDir.hh"
#include "Log.hh"
#include "MainWindow.hh"
#include "MeshHistory.hh"
#include "ModelPreview.hh"
#include "Mirror.hh"
#include "Move.hh"
#include "Preferences.hh"
#include "SaveDialog.hh"
#include "Scale.hh"
#include "Tools.hh"

#ifdef WIN32
#undef OUT
#undef IN
#endif

#include <gtkmm/aboutdialog.h>
#include <gtkmm/checkbutton.h>
#include <gtkmm/colorbutton.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/frame.h>
#include <gtkmm/icontheme.h>
#include <gtkmm/main.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/notebook.h>
#include <gtkmm/radioaction.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/scale.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/statusbar.h>
#include <gtkmm/stock.h>
#include <gtkmm/toggleaction.h>
#include <gtkmm/togglebutton.h>
#include <gtkmm/treeview.h>
#include <sys/stat.h>
#include <cmath>

#include <iostream>

using SharpConstruct::GInterface::MainWindow;
using SharpConstruct::Progress;

MainWindow::MainWindow()
: deformations_( progress_ )
{
#ifndef WIN32
	gbr_init( 0 );
#endif

	editor_.SignalExtensions().connect( sigc::mem_fun( *this, &MainWindow::on_editor_extensions_ ) );

	const Glib::ustring path( DataDir( "/sharpconstruct/interface/" ) );

	Gtk::IconTheme::get_default()->append_search_path( path + "pixmaps/" );

	// Workaround for Bug #1483661
	std::string buffer = Glib::file_get_contents( path + "SharpConstruct.glade" );
	glade_ = Gnome::Glade::Xml::create_from_buffer( buffer.c_str(), buffer.length() );

	// Need to set about dialog logo manually
	Glib::RefPtr<Gdk::Pixbuf> aboutDialogLogo = Gdk::Pixbuf::create_from_file( DataDir( "/sharpconstruct/interface/logo.png" ) );
	Gtk::AboutDialog* aboutDialog;
	glade_->get_widget( "AboutDialog", aboutDialog );
	aboutDialog->set_logo ( aboutDialogLogo );

	// Put the editor into the interface
	Gtk::VBox* modelbox;
	glade_->get_widget( "ModelBox", modelbox );
	modelbox->add( editor_ );
	glade_->get_widget( "BrushModifiers", modelbox );
	modelbox->pack_end( Tools::Instance().BrushCurve() );

	// Must go before signals are connected!
	const int gm( Editor::DisplayOptionsData::VBO );
	if( Preferences::Instance().Get( "GraphicsMode", gm ) ==
	    Editor::DisplayOptionsData::VAR )
	{
		Gtk::RadioButton* rb;
		glade_->get_widget( "VAR", rb );
		rb->set_active( true );
	}

	icon_factory_ = Gtk::IconFactory::create();

	init_actions_();
	connect_handlers_();
	init_brushes_();
}

void MainWindow::Run( Gtk::Main& kit )
{
	Gtk::Window* mainwindow( 0 );
	glade_->get_widget( "MainWindow", mainwindow );
	initialize_();
	create_model_();

	kit.run( *mainwindow );
}

void MainWindow::add_stock_item_( const std::string& filepath,
				  const Glib::ustring& id, const Glib::ustring& label,
				  const Gdk::ModifierType mod, const unsigned key )
{
	Gtk::IconSource source;
	try
	{
		source.set_pixbuf( Gdk::Pixbuf::create_from_file(filepath) );
	}
	catch( const Glib::Exception& ex )
	{
		std::cout << ex.what() << std::endl;
	}

	source.set_size( Gtk::ICON_SIZE_SMALL_TOOLBAR );
	source.set_size_wildcarded(); //Icon may be scaled.

	Gtk::IconSet icon_set;
	icon_set.add_source( source ); //More than one source per set is allowed.

	const Gtk::StockID stock_id( id );
	icon_factory_->add( stock_id, icon_set );

	if( key == '\0' )
		Gtk::Stock::add( Gtk::StockItem( stock_id, label ) );
	else
		Gtk::Stock::add( Gtk::StockItem( stock_id, label, mod, key ) );
}

void MainWindow::add_stock_item_( const std::string& filepath,
				  const Glib::ustring& id, const Glib::ustring& label )
{
	add_stock_item_( filepath, id, label, Gdk::ModifierType(), '\0' );
}

void MainWindow::init_actions_()
{
	using sigc::mem_fun;

	Gtk::AboutDialog* ad;
	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );
	glade_->get_widget( "AboutDialog", ad );
	// Create new stock items
	{
		const std::string path( DataDir( "/sharpconstruct/interface/" ) );
		add_stock_item_( path + "displacebrush.png", "SC_DISPLACE_BRUSH", "_Draw" );
		add_stock_item_( path + "grabbrush.png", "SC_GRAB_BRUSH", "_Grab" );
		add_stock_item_( path + "pinchbrush.png", "SC_PINCH_BRUSH", "_Pinch" );
		add_stock_item_( path + "smoothbrush.png", "SC_SMOOTH_BRUSH", "_Smooth" );
		add_stock_item_( path + "wire.png", "SC_WIRE", "_Wire", Gdk::CONTROL_MASK, 'w' );
		add_stock_item_( path + "smoothed.png", "SC_SMOOTH", "_Smooth" );
		add_stock_item_( path + "double.png", "SC_DOUBLE", "_Double" );
		add_stock_item_( path + "speed.png", "SC_SPEED", "S_peed" );
		add_stock_item_( path + "center.png", "SC_CENTER", "_Center", Gdk::CONTROL_MASK, 'c' );
		icon_factory_->add_default();
	}

	action_group_ = Gtk::ActionGroup::create();
	action_group_->add( Gtk::Action::create( "MenuFile", "_File" ) );
	action_group_->add( Gtk::Action::create( "New", Gtk::Stock::NEW, "", "Create new model" ),
			    mem_fun( *this, &MainWindow::new_model_ ) );
	action_group_->add( Gtk::Action::create( "Open", Gtk::Stock::OPEN, "", "Open model" ),
			    mem_fun( *this, &MainWindow::open_model_ ) );
	action_group_->add( Gtk::Action::create( "Save", Gtk::Stock::SAVE, "", "Save model" ),
			    mem_fun( *this, &MainWindow::save_model_ ) );
	action_group_->add( Gtk::Action::create( "SaveAs", Gtk::Stock::SAVE_AS, "", "Save model under new name" ),
			    mem_fun( *this, &MainWindow::save_model_as_ ) );
	action_group_->add( Gtk::Action::create( "Quit", Gtk::Stock::QUIT ),
			    mem_fun( *this, &MainWindow::on_quit_ ) );
	action_group_->add( Gtk::Action::create( "MenuEdit", "_Edit" ) );
	action_group_->add( Gtk::Action::create( "Undo", Gtk::Stock::UNDO ), Gtk::AccelKey( "<control>z", "<Actions>//Undo" ),
			    mem_fun( *this, &MainWindow::on_undo_clicked_ ) );
	action_group_->add( Gtk::Action::create( "Redo", Gtk::Stock::REDO ), Gtk::AccelKey( "<control><shift>z", "<Actions>//Redo" ),
			    mem_fun( *this, &MainWindow::on_redo_clicked_ ) );
	action_group_->add( Gtk::Action::create( "MenuBrush", "_Brush" ) );
	Gtk::RadioAction::Group brush_group;
	action_group_->add( Gtk::RadioAction::create( brush_group, "Draw", Gtk::StockID( "SC_DISPLACE_BRUSH" ) ), Gtk::AccelKey( "q", "<Actions>//Draw" ),
			    mem_fun( *this, &MainWindow::on_draw_clicked_ ) );
	action_group_->add( Gtk::RadioAction::create( brush_group, "Smooth", Gtk::StockID( "SC_SMOOTH_BRUSH" ) ), Gtk::AccelKey( "w", "<Actions>//Smooth" ),
			    mem_fun( *this, &MainWindow::on_smooth_clicked_ ) );
	action_group_->add( Gtk::RadioAction::create( brush_group, "Pinch", Gtk::StockID( "SC_PINCH_BRUSH" ) ), Gtk::AccelKey( "e", "<Actions>//Pinch" ),
			    mem_fun( *this, &MainWindow::on_pinch_clicked_ ) );
	action_group_->add( Gtk::RadioAction::create( brush_group, "Grab", Gtk::StockID( "SC_GRAB_BRUSH" ) ), Gtk::AccelKey( "r", "<Actions>//Grab" ),
			    mem_fun( *this, &MainWindow::on_grab_clicked_ ) );
	action_group_->add( Gtk::Action::create( "IncreaseSize", "_Increase Size" ), Gtk::AccelKey( "f", "<Actions>//IncreaseSize" ),
			    mem_fun( *this, &MainWindow::on_increase_size_clicked_ ) );
	action_group_->add( Gtk::Action::create( "DecreaseSize", "_Decrease Size" ), Gtk::AccelKey( "d", "<Actions>//DecreaseSize" ),
			    mem_fun( *this, &MainWindow::on_decrease_size_clicked_ ) );
	action_group_->add( Gtk::Action::create( "IncreaseStrength", "_Increase Strength" ), Gtk::AccelKey( "s", "<Actions>//IncreaseStrength" ),
			    mem_fun( *this, &MainWindow::on_increase_strength_clicked_ ) );
	action_group_->add( Gtk::Action::create( "DecreaseStrength", "_Decrease Strength" ), Gtk::AccelKey( "a", "<Actions>//DecreaseStrength" ),
			    mem_fun( *this, &MainWindow::on_decrease_strength_clicked_ ) );
	action_group_->add( Gtk::Action::create( "MenuView", "_View" ) );
	action_group_->add( Gtk::ToggleAction::create( "ShowSidebar", "Show Side_bar" ), Gtk::AccelKey( "F8", "<Actions>//ShowSidebar" ),
			    mem_fun( *this, &MainWindow::on_sidebar_toggled_ ) );
	action_group_->add( Gtk::ToggleAction::create( "SidebarSide", "Sidebar on _Right" ), Gtk::AccelKey( "F9", "<Actions>//SidebarSide" ),
			    mem_fun( *this, &MainWindow::on_sidebar_side_toggled_ ) );
	action_group_->add( Gtk::ToggleAction::create( "Wireframe", Gtk::StockID( "SC_WIRE" ), "", "Overlay polygon wireframe" ),
			    mem_fun( editor_.DisplayOptions(), &Editor::DisplayOptionsData::ToggleWireframe ) );
	action_group_->add( Gtk::ToggleAction::create( "SmoothView", Gtk::StockID( "SC_SMOOTH" ), "", "Smooth polygon edges" ),
			    mem_fun( editor_.DisplayOptions(), &Editor::DisplayOptionsData::ToggleSmoothed ) );
	action_group_->add( Gtk::ToggleAction::create( "Double", Gtk::StockID( "SC_DOUBLE" ), "", "Draw back side of polygons" ),
			    mem_fun( editor_.DisplayOptions(), &Editor::DisplayOptionsData::ToggleDoubleSided ) );
	action_group_->add( Gtk::ToggleAction::create( "Speed", Gtk::StockID( "SC_SPEED" ),"", "Restrict model redraw to increase speed, may cause redraw artifacts" ),
			    mem_fun( editor_.DisplayOptions(), &Editor::DisplayOptionsData::ToggleSpeed ) );
	action_group_->add( Gtk::Action::create( "Side", "Side" ), Gtk::AccelKey( "KP_3", "<Actions>//Side" ),
			    mem_fun( editor_, &Editor::SideView ) );
	action_group_->add( Gtk::Action::create( "Front", "Front" ), Gtk::AccelKey( "KP_1", "<Actions>//Front" ),
			    mem_fun( editor_, &Editor::FrontView ) );
	action_group_->add( Gtk::Action::create( "Top", "Top" ), Gtk::AccelKey( "KP_7", "<Actions>//Top" ),
			    mem_fun( editor_, &Editor::TopView ) );
	action_group_->add( Gtk::Action::create( "Center", Gtk::StockID( "SC_CENTER" ) ),
			    mem_fun( editor_, &Editor::ResetView ) );
	action_group_->add( Gtk::Action::create( "ZoomIn", "Zoom In" ), Gtk::AccelKey( "KP_Add", "<Actions>//ZoomIn" ),
			    mem_fun( *this, &MainWindow::zoom_in_ ) );
	action_group_->add( Gtk::Action::create( "ZoomOut", "Zoom Out" ), Gtk::AccelKey( "KP_Subtract", "<Actions>//ZoomOut" ),
			    mem_fun( *this, &MainWindow::zoom_out_ ) );

        Gtk::RadioAction::Group shader_group;
	action_group_->add( Gtk::RadioAction::create( shader_group, "ShaderPlain", "Plain shading" ),
			    mem_fun( *this, &MainWindow::shader_plain_ ) );
	action_group_->add( Gtk::RadioAction::create( shader_group, "ShaderNormals", "Normals shading" ),
			    mem_fun( *this, &MainWindow::shader_normals_ ) );
        if( !editor_.ShadersAvailable() )
        {
		action_group_->get_action( "ShaderPlain" )->set_sensitive( false );
		action_group_->get_action( "ShaderNormals" )->set_sensitive( false );
        }

	action_group_->add( Gtk::Action::create( "MenuDeformations", "_Deformations" ) );
	action_group_->add( Gtk::Action::create( "Subdivide", "_Subdivide" ), Gtk::AccelKey( "<alt>s", "<Actions>//Subdivide" ),
			    mem_fun( deformations_, &Deformations::Subdivide ) );
	action_group_->add( Gtk::Action::create( "FlipNormals", "_Flip Normals" ),
			    mem_fun( deformations_, &Deformations::FlipNormals ) );
	action_group_->add( Gtk::Action::create( "Unify", "_Unify" ), Gtk::AccelKey( "<alt>u", "<Actions>//Unify" ),
			    mem_fun( deformations_, &Deformations::Unify ) );
	action_group_->add( Gtk::Action::create( "ToTriangles", "_To Triangles" ),
			    mem_fun( deformations_, &Deformations::ToTriangles ) );
	action_group_->add( Gtk::Action::create( "MenuHelp", "_Help" ) );
	action_group_->add( Gtk::Action::create( "About", Gtk::Stock::ABOUT ),
			    mem_fun( *ad, &Gtk::AboutDialog::show ) );

	// Configure actions
	Glib::RefPtr< Gtk::ToggleAction >::cast_dynamic( action_group_->get_action( "ShowSidebar" ) )->set_active( true );
	Glib::RefPtr< Gtk::ToggleAction >::cast_dynamic( action_group_->get_action( "SmoothView" ) )->set_active( true );
	Glib::RefPtr< Gtk::ToggleAction >::cast_dynamic( action_group_->get_action( "Double" ) )->set_active( true );

	ui_manager_ = Gtk::UIManager::create();
	ui_manager_->insert_action_group( action_group_ );

	mainwindow->add_accel_group( ui_manager_->get_accel_group() );
	ui_manager_->add_ui_from_file( DataDir( "/sharpconstruct/interface/UIManager.xml" ) );

	Gtk::VBox* vb;
	glade_->get_widget( "MainVBox", vb );
	vb->pack_start( *ui_manager_->get_widget( "/MenuBar" ), Gtk::PACK_SHRINK );
	vb->pack_start( *ui_manager_->get_widget( "/ToolBar" ), Gtk::PACK_SHRINK );
	vb->reorder_child( *ui_manager_->get_widget( "/ToolBar" ), 0 );
	vb->reorder_child( *ui_manager_->get_widget( "/MenuBar" ), 0 );
}

void MainWindow::connect_handlers_()
{
	using sigc::mem_fun;

	Gtk::Widget* w;
	Gtk::CheckButton* chb;
	Gtk::ColorButton* cb;
	Gtk::HScale* hs;
	Gtk::Scale* scale;
	Gtk::TreeView* iv;
	Gtk::RadioButton* rb;
	Gtk::SpinButton* sb;
	Gtk::ToggleButton* tgb;

	using namespace sigc;
	MeshHistory::Instance().Changed().
		connect( hide( hide( mem_fun( *this, 
					      &MainWindow::
					      on_history_changed_ ) ) ) );
	MeshHistory::Instance().SignalEntireSectionChanged().connect( hide( hide( hide( sigc::mem_fun( *this, &MainWindow::update_model_info_ ) ) ) ) );

	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );
	mainwindow->signal_delete_event().connect( mem_fun( *this, &MainWindow::on_delete_event_ ) );

	// Connect brush modifiers

	// Stroke
	glade_->get_widget( "Layer", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "Dot", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "MaskBrush", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "Radius", hs );
	hs->signal_value_changed().connect( mem_fun( *this, &MainWindow::on_radius_changed_ ) );
	glade_->connect_clicked( "SizePressure", mem_fun( *this, &MainWindow::on_radius_pressure_changed_ ) );
	glade_->get_widget( "Strength", hs );
	hs->signal_value_changed().connect( mem_fun( *this, &MainWindow::on_strength_changed_ ) );
	glade_->connect_clicked( "StrengthPressure", mem_fun( *this, &MainWindow::on_strength_pressure_changed_ ) );
	glade_->get_widget( "Intensity", hs );
	hs->signal_value_changed().connect( mem_fun( *this, &MainWindow::on_intensity_changed_ ) );
	glade_->connect_clicked( "IntensityPressure", mem_fun( *this, &MainWindow::on_intensity_pressure_changed_ ) );

	glade_->get_widget( "BrushAdd", rb );
	rb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_direction_ ) );
	glade_->get_widget( "BrushSub", rb );
	rb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_direction_ ) );
	glade_->get_widget( "BrushNoDirection", rb );
	rb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_direction_ ) );
	glade_->get_widget( "Colorize", chb );
	chb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "BrushColor", cb );
	cb->signal_color_set().connect( mem_fun( *this, &MainWindow::update_brush_ ) );

	glade_->get_widget( "SymmetryX", chb );
	chb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "SymmetryY", chb );
	chb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );
	glade_->get_widget( "SymmetryZ", chb );
	chb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_brush_ ) );

	glade_->get_widget( "BrushShapes", iv );
	iv->get_selection()->signal_changed().connect( mem_fun( *this, &MainWindow::update_brush_shape_ ) );
	glade_->connect_clicked( "BrushShapeAdd", mem_fun( *this, &MainWindow::on_brush_shape_add_clicked_ ) );
	glade_->connect_clicked( "BrushShapeRemove", mem_fun( *this, &MainWindow::on_brush_shape_remove_clicked_ ) );

	glade_->get_widget( "EnableBrushCurve", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::on_enable_brush_curve_toggled_ ) );

	glade_->get_widget( "InteractiveOff", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_interactive_deformation_ ) );
	glade_->get_widget( "Mirror", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_interactive_deformation_ ) );
	glade_->get_widget( "Move", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_interactive_deformation_ ) );
	glade_->get_widget( "Scale", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::update_interactive_deformation_ ) );
	glade_->connect_clicked( "InteractiveApply", mem_fun( *this, &MainWindow::apply_interactive_deformation_ ) );

	// Connect settings
	glade_->get_widget( "BackgroundColor", cb );
	cb->signal_color_set().connect( mem_fun( *this, &MainWindow::on_background_color_set_ ) );
	glade_->get_widget( "UndoSteps", sb );
	sb->signal_value_changed().connect( mem_fun( *this, &MainWindow::on_undo_steps_value_changed_ ) );
	glade_->get_widget( "VBO", rb );
	glade_->connect_clicked( "VBO", mem_fun( *this, &MainWindow::update_graphics_mode_settings_ ) );
	glade_->connect_clicked( "SpeedTest", mem_fun( *this, &MainWindow::on_speed_test_clicked_ ) );
	glade_->get_widget( "SimpleRotation", tgb );
	tgb->signal_toggled().connect( mem_fun( *this, &MainWindow::on_simple_rotation_toggled_ ) );
	glade_->connect_clicked( "InputDevices", mem_fun( input_dialog_, &InputDialog::Show ) );

	// Lighting settings
	glade_->get_widget( "DiffuseColor", cb );
	cb->signal_color_set().connect( mem_fun( *this, &MainWindow::update_lighting_ ) );
	glade_->get_widget( "SpecularColor", cb );
	cb->signal_color_set().connect( mem_fun( *this, &MainWindow::update_lighting_ ) );
	glade_->get_widget( "Shininess", scale );
	scale->signal_value_changed().connect( mem_fun( *this, &MainWindow::update_lighting_ ) );

	glade_->get_widget( "ExtError", w );
	glade_->connect_clicked( "ExtErrorOK", mem_fun( w, &Gtk::Widget::hide ) );

	glade_->get_widget( "Zoom", scale );
	scale->signal_value_changed().connect( mem_fun( *this, &MainWindow::update_zoom_ ) );
}

void MainWindow::initialize_()
{
	Gtk::AboutDialog* ad;
	Gtk::ColorButton* cb;
	Gtk::ToggleButton* tgb;
	Gtk::Window* mainwindow;
	Gtk::SpinButton* sb;
	glade_->get_widget( "AboutDialog", ad );
	ad->set_version( GetVersion() );
	input_dialog_.Initialize();
	deformations_.Initialize( glade_ );
	mask_.Initialize( glade_ );
	glade_->get_widget( "BrushColor", cb );
	cb->set_color( Gdk::Color( "White" ) );
	{
		glade_->get_widget( "BackgroundColor", cb );
		Gdk::Color c;
		c.set_rgb_p( Preferences::Instance().Get( "VIEW_COLOR_R", 0.3 ),
			     Preferences::Instance().Get( "VIEW_COLOR_G", 0.3 ),
			     Preferences::Instance().Get( "VIEW_COLOR_B", 0.3 ) );
		cb->set_color( c );
	}
	glade_->get_widget( "UndoSteps", sb );
	sb->set_value( Preferences::Instance().Get( "UNDO_STEPS", 3 ) );
	glade_->get_widget( "MainWindow", mainwindow );
	if( !Preferences::Instance().Get( "Maximized", false ) )
	{
		mainwindow->resize( Preferences::Instance().Get( "W", 800 ),
				    Preferences::Instance().Get( "H", 600 ) );
	}
	else
		mainwindow->maximize();
#ifndef WIN32
	mainwindow->set_icon_from_file( DataDir( "/pixmaps/sharpconstruct-icon.png" ) );
#endif

	{
		const std::string path( DataDir( "/sharpconstruct/interface/" ) );
		add_stock_item_( path + "pressure.png", "SC_PRESSURE", "" );
		glade_->get_widget( "SizePressure", tgb );
		tgb->set_label( "SC_PRESSURE" );
		tgb->set_use_stock( true );
		glade_->get_widget( "StrengthPressure", tgb );
		tgb->set_label( "SC_PRESSURE" );
		tgb->set_use_stock( true );
		glade_->get_widget( "IntensityPressure", tgb );
		tgb->set_label( "SC_PRESSURE" );
		tgb->set_use_stock( true );
	}

	// Add progress bar to the status bar
	{
		Gtk::Statusbar* sbar;
		glade_->get_widget( "Statusbar", sbar );
		//sbar->pack_end( progress_bar_ );
		//progress_bar_.show();
		//progress_.ChangedSignal().connect( sigc::mem_fun( *this, &MainWindow::update_progress_bar_ ) );
	}

	Gtk::Scale* scale;
	glade_->get_widget( "Zoom", scale );
	editor_.SignalZoomChanged().connect( sigc::mem_fun( *scale, &Gtk::VScale::set_value ) );
	
	glade_->get_widget( "SimpleRotation", tgb );
	tgb->set_active( Preferences::Instance().Get( "SimpleRotation", false ) );

	glade_->get_widget( "DiffuseColor", cb );
	cb->set_color( editor_.Lighting().Diffuse().GdkColor() );
	glade_->get_widget( "SpecularColor", cb );
	cb->set_color( editor_.Lighting().Specular().GdkColor() );
}

void MainWindow::update_model_info_()
{
	Gtk::Label* l;
	glade_->get_widget( "ModelLabel", l );
	const Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );
	std::string file( Glib::filename_display_basename( mesh.FileName() ) );
	if( file == "." )
		file = "untitled";
	l->set_text( file + ": " + mesh.GetModelStatisticsString() );
}

void MainWindow::init_brushes_()
{
	update_brush_();

	Gtk::TreeView* iv;
	glade_->get_widget( "BrushShapes", iv );

	brush_shape_model_ = Gtk::ListStore::create( brush_shape_columns_ );
	brush_shape_model_->clear();

	std::string def_path;
	def_path = DataDir( "/sharpconstruct/brushes/" );

	{
		Gtk::TreeView::Column* pColumn = Gtk::manage( new Gtk::TreeView::Column( "" ) );

		pColumn->pack_start( brush_shape_columns_.Icon, false );
		pColumn->pack_start( brush_shape_columns_.Name );

		iv->append_column( *pColumn );
	}

	Glib::Dir dir( def_path );
	std::string name( dir.read_name() );
	Gtk::TreeModel::iterator def;
	while( !name.empty() )
	{
		if( name[ 0 ] != '.' )
		{
			const std::string path( Glib::build_filename( def_path,
								      name ) );
			Glib::ustring display_name
				( Glib::filename_to_utf8( name ) );

			Gtk::TreeModel::iterator iter;
			try
			{
				iter = brush_shape_model_->append();
				Gtk::TreeModel::Row row = *iter;

				const unsigned extdot( display_name.find( "." ) );
				if( extdot != std::string::npos )
					display_name = display_name.substr( 0, extdot );
				if( display_name == "Default" )
					def = iter;
				row[ brush_shape_columns_.Name ] = display_name;
				row[ brush_shape_columns_.Icon ] =
					Gdk::Pixbuf::create_from_file( path )->
					scale_simple( 32, 32,
						      Gdk::INTERP_BILINEAR );
				row[ brush_shape_columns_.Shape ] =
					Gdk::Pixbuf::create_from_file( path );
			}
			catch( const Gdk::PixbufError& )
			{
				brush_shape_model_->erase( iter );
			}
		}
		
		name = dir.read_name();
	}

	brush_shape_model_->move( def, brush_shape_model_->children().begin() );

	iv->set_model( brush_shape_model_ );

	iv->get_selection()->set_mode( Gtk::SELECTION_BROWSE );
	iv->get_selection()->select( brush_shape_model_->get_path( def ) );

	Gtk::ListStore::iterator iter( iv->get_selection()->get_selected() );
	if( iter )
	{
		brush_shape_ = ( *iter )[ brush_shape_columns_.Shape ];
		Tools::Instance().SetInitialBrushShape( brush_shape_, brush_shape_model_->get_path( def ) );
	}
}

void MainWindow::create_model_()
{
	Gtk::Notebook* mdltype;
	glade_->get_widget( "ModelType", mdltype );
	const Glib::ustring type( mdltype->get_current()->get_tab_label_text() );

	MeshHistory::Instance().Clear();
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	Gtk::SpinButton* sb;
	if( type == "Cube" )
	{
		glade_->get_widget( "CubeX", sb );
		const unsigned x( sb->get_value_as_int() + 1 );
		glade_->get_widget( "CubeY", sb );
		const unsigned y( sb->get_value_as_int() + 1 );
		glade_->get_widget( "CubeZ", sb );
		const unsigned z( sb->get_value_as_int() + 1 );

		mesh.Create( Create::Cube( progress_, x, y, z ), progress_ );
	}
	else if( type == "Plane" )
	{
		glade_->get_widget( "PlaneX", sb );
		const unsigned x( sb->get_value_as_int() + 1 );
		glade_->get_widget( "PlaneY", sb );
		const unsigned y( sb->get_value_as_int() + 1 );

		mesh.Create( Create::Plane( x, y ), progress_ );
	}
	else if( type == "Sphere" )
	{
		glade_->get_widget( "SphereDiv", sb );
		const unsigned div( sb->get_value_as_int() + 1 );

		mesh.Create( Create::Sphere( progress_, div ), progress_ );
		progress_.Update( 0.75, "outside" );
	}

	on_history_changed_(); // For save btn
	editor_.ResetView();
	editor_.Refresh( true, true );
}

bool MainWindow::is_mesh_data_safe_()
{
	const bool SAFE = true;
	const bool UNSAFE = false;
	
	if( MeshHistory::Instance().Saved() )
		return SAFE;
	else
	{
		Gtk::Window* mainwindow;
		glade_->get_widget( "MainWindow", mainwindow );
		
		Gtk::MessageDialog md( "Save changes to model before closing? ",
				       false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_NONE );
		md.set_title( "" );
		md.set_secondary_text( "If you don't save the model, unsaved changes will be lost." );
		md.set_transient_for( *mainwindow );

		md.add_button( "Discard", Gtk::RESPONSE_NO );
		md.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
		md.add_button( Gtk::Stock::SAVE, Gtk::RESPONSE_YES );

		const int rsp( md.run() );

		if( rsp == Gtk::RESPONSE_NO )
			return SAFE;
		else if( rsp == Gtk::RESPONSE_CANCEL )
			return UNSAFE;
		else if( rsp == Gtk::RESPONSE_YES )
		{
			save_model_();
			if( MeshHistory::Instance().Saved() )
				return SAFE;
			else
				return UNSAFE;
		}
		else
			return UNSAFE;
	}
}

void MainWindow::on_background_color_set_()
{
	Gtk::ColorButton* cb;
	glade_->get_widget( "BackgroundColor", cb );
	const Gdk::Color c( cb->get_color() );
	editor_.DisplayOptions().SetBackgroundColor( Color( c.get_red_p(),
							    c.get_green_p(),
							    c.get_blue_p() ) );
}

void MainWindow::on_brush_shape_add_clicked_()
{
	Gtk::TreeView* tv;
	glade_->get_widget( "BrushShapes", tv );

	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );

	Gtk::FileChooserDialog fd( "Add Brush Shape" );
	fd.set_transient_for( *mainwindow );
	fd.set_select_multiple( true );

	fd.set_action( Gtk::FILE_CHOOSER_ACTION_OPEN );
	fd.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
	fd.add_button( Gtk::Stock::OPEN, Gtk::RESPONSE_OK );

	{
		Gtk::FileFilter filter;
		filter.set_name( "All Images" );
		filter.add_pixbuf_formats();
		fd.add_filter( filter );
	}
	
	{
		Gtk::FileFilter filter;
		filter.set_name( "All files" );
		filter.add_pattern( "*" );
		fd.add_filter( filter );
	}

	const int result( fd.run() );

	if( result == Gtk::RESPONSE_OK )
	{
		Glib::SListHandle< Glib::ustring > files( fd.get_filenames() );
		Glib::ustring name;
		for( Glib::SListHandle< Glib::ustring >::iterator i = files.begin();
		     i != files.end(); ++i )
		{
			Gtk::ListStore::iterator cur( tv->get_selection()->get_selected() ), iter;
			iter = brush_shape_model_->insert_after( cur );

			try
			{
				Gtk::TreeModel::Row row = *iter;
				
				name = path_get_basename( *i );
				const unsigned extdot( name.find( "." ) );
				if( extdot != std::string::npos )
					name = name.substr( 0, extdot );
				row[ brush_shape_columns_.Name ] = name;
				Glib::RefPtr< Gdk::Pixbuf > input( Gdk::Pixbuf::create_from_file( *i ) );
				if( input->get_width() != input->get_height() )
				{
					input = input->scale_simple( input->get_width(),
								     input->get_width(),
								     Gdk::INTERP_BILINEAR );
				}
				row[ brush_shape_columns_.Icon ] =
					input->scale_simple( 32, 32,
							     Gdk::INTERP_BILINEAR );
				row[ brush_shape_columns_.Shape ] = input;
			}
			catch( const Gdk::PixbufError& )
			{
				brush_shape_model_->erase( iter );
			}
		}
	}
}
void MainWindow::on_brush_shape_remove_clicked_()
{
	Gtk::TreeView* tv;
	glade_->get_widget( "BrushShapes", tv );

	if( brush_shape_model_->children().size() > 1 )
	{
		Gtk::ListStore::iterator iter( tv->get_selection()->get_selected() ),
			after( iter ), before( iter );
		after++;
		if( iter != brush_shape_model_->children().begin() )
			before--;
		brush_shape_model_->erase( iter++ );
		if( after != brush_shape_model_->children().end() )
			tv->get_selection()->select( after );
		else
			tv->get_selection()->select( before );
	}
}

void MainWindow::on_decrease_size_clicked_()
{
	Gtk::HScale* hs;
	glade_->get_widget( "Radius", hs );
	hs->set_value( hs->get_value() - 2 );
}

void MainWindow::on_decrease_strength_clicked_()
{
	Gtk::HScale* hs;
	glade_->get_widget( "Strength", hs );
	hs->set_value( hs->get_value() - 2 );
}

bool MainWindow::on_delete_event_( GdkEventAny* )
{
	if( is_mesh_data_safe_() )
	{
		Gtk::SpinButton* sb;
		Gtk::Window* mw;
		glade_->get_widget( "UndoSteps", sb );
		Preferences::Instance().Set( "UNDO_STEPS", (int)sb->get_value() );
		glade_->get_widget( "MainWindow", mw );
		Preferences::Instance().Set( "W", mw->get_width() );
		Preferences::Instance().Set( "H", mw->get_height() );
		const Gdk::WindowState ws( mw->get_window()->get_state() );
		if( ws & Gdk::WINDOW_STATE_MAXIMIZED )
			Preferences::Instance().Set( "Maximized", true );
		else
			Preferences::Instance().Set( "Maximized", false );

#ifdef _MSC_VER
		// This is a temporary hack to prevent the crash-at-exit on Win32
                exit(0);
#endif
		return false;
	}
	else
        {
		return true;
        }
}

void MainWindow::on_editor_extensions_( bool vbo, bool vs )
{
	Gtk::Widget* w;

	// For testing:
	//vbo = false;
	//vs = false;
	if( !vbo || !vs )
	{
		glade_->get_widget( "ExtError", w );
		w->show();

		std::string err( "Your graphics card or drivers do not support the following extension" );
		if( !vbo && !vs )
			err += "s";
		err += ": ";
		std::string exts;
		if( !vbo )
			exts += "ARB Vertex Buffer Objects";
		if( exts.length() )
			exts += ", ";
		if( !vs )
			exts += "ARB Shader Objects";
		err += exts;

		Gtk::Label* lbl;
		glade_->get_widget( "ExtErrorText", lbl );
		lbl->set_label( err );
	}

	if( !vbo )
	{
		glade_->get_widget( "VBO", w );
		w->set_sensitive( false );
	}

	if( !vbo || Preferences::Instance().Get( "GraphicsMode", 0 ) == 1 )
	{
		Gtk::ToggleButton* tgb;
		glade_->get_widget( "VAR", tgb );
		tgb->set_active( true );
	}

	// Display an error if the shader log is not empty
	if( editor_.ShaderLog().length() )
	{
		Gtk::Window* mainwindow;
		glade_->get_widget( "MainWindow", mainwindow );
		
		Gtk::MessageDialog md( "The vertex shader contains an error", false, Gtk::MESSAGE_WARNING );
		md.set_title( "" );
		md.set_secondary_text( editor_.ShaderLog() );
		md.set_transient_for( *mainwindow );
		
		md.run();
	}
}

void MainWindow::on_history_changed_()
{
	action_group_->get_action( "Save" )->set_sensitive( !MeshHistory::Instance().Saved() );
	action_group_->get_action( "Undo" )->set_sensitive( MeshHistory::Instance().UndoPossible() );
	action_group_->get_action( "Redo" )->set_sensitive( MeshHistory::Instance().RedoPossible() );

	update_model_info_();
}

void MainWindow::on_increase_size_clicked_()
{
	Gtk::HScale* hs;
	glade_->get_widget( "Radius", hs );
	hs->set_value( hs->get_value() + 2 );
}

void MainWindow::on_increase_strength_clicked_()
{
	Gtk::HScale* hs;
	glade_->get_widget( "Strength", hs );
	hs->set_value( hs->get_value() + 2 );
}

void MainWindow::on_undo_clicked_()
{
	MeshHistory::Instance().Undo();
	editor_.Refresh( true, true );
}

void MainWindow::on_redo_clicked_()
{
	MeshHistory::Instance().Redo();
	editor_.Refresh( true, true );
}

void MainWindow::on_quit_()
{
	if( is_mesh_data_safe_() )
	{
		Gtk::Window* mw;
		glade_->get_widget( "MainWindow", mw );
		mw->hide();
#ifdef _MSC_VER
		// This is a hack to prevent the crash-at-exit on Win32
                exit(0);
#endif
	}
}

void MainWindow::on_sidebar_toggled_()
{
	Gtk::Widget* w;

	glade_->get_widget( "Sidebar", w );

	const bool v( Glib::RefPtr< Gtk::ToggleAction >::cast_dynamic( action_group_->get_action( "ShowSidebar" ) )->get_active() );
	if( v )
		w->show();
	else
		w->hide();
}

void MainWindow::on_sidebar_side_toggled_()
{
	Gtk::Box* b;
	Gtk::Widget* w;
	glade_->get_widget( "MainHBox", b );
	glade_->get_widget( "Sidebar", w );
	const bool v( Glib::RefPtr< Gtk::ToggleAction >::cast_dynamic( action_group_->get_action( "SidebarSide" ) )->get_active() );
	if( v )
		b->reorder_child( *w, 1 );
	else
		b->reorder_child( *w, 0 );
	editor_.Refresh( false, false );
}

void MainWindow::on_simple_rotation_toggled_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "SimpleRotation", tgb );
	editor_.DisplayOptions().SetSimpleRotation( tgb->get_active() );
	Preferences::Instance().Set( "SimpleRotation", tgb->get_active() );
}

void MainWindow::on_speed_test_clicked_()
{
	const int test( 0 );

	Gtk::Window* mw;
	glade_->get_widget( "MainWindow", mw );

	if( test == 0 )
	{
		const unsigned runs( 10 );
		
		Gtk::Statusbar* sb;
		glade_->get_widget( "Statusbar", sb );
		
		std::vector< double > times;
		std::string output;
		double adder( 0 );
		for( unsigned i = 0; i < runs; ++i )
		{
			MeshHistory::Instance().AddMesh( true, false, false );
			Glib::Timer timer;
			MeshHistory::Instance().GetCurrentMesh().PerformanceTest( "" );
			timer.stop();
			times.push_back( timer.elapsed() );
			output += ToString( times[ i ] ) + " seconds\n";
			adder += times[ i ];
		}
		Gtk::MessageDialog::MessageDialog( *mw, "Times:\n" + output +
						   "\nAverage: " + ToString( adder / runs ) +
						   " seconds" ).run();
	}
	else if( test == 1 )
	{
		GdkEventButton click;
		//GdkEventMotion em;

		int x( editor_.get_width() / 2 + editor_.get_allocation().get_x() );
		int y( editor_.get_height() / 2 + editor_.get_allocation().get_y() );

		click.type = GDK_BUTTON_PRESS;
		click.window = editor_.get_window()->gobj();
		click.x = x;
		click.y = y;
		click.button = 1;
		click.device = ( *editor_.get_display()->list_devices().begin() )->gobj();

		const unsigned runs( 100 );
		Glib::Timer timer;
		for( unsigned i = 0; i < runs; ++i )
			editor_.event( (GdkEvent*)( &click ) );
		timer.stop();
		Gtk::MessageDialog::MessageDialog( *mw, ToString( timer.elapsed() ) + " seconds" ).run();	

		/*em.type = GDK_MOTION_NOTIFY;
		em.window = editor_.get_window()->gobj();
		em.x = 200;
		em.y = 200;
		em.device = ( *editor_.get_display()->list_devices().begin() )->gobj();
		editor_.event( (GdkEvent*)( &em ) );*/
	}
}

void MainWindow::on_undo_steps_value_changed_()
{
	Gtk::SpinButton* sb;
	glade_->get_widget( "UndoSteps", sb );
	MeshHistory::Instance().SetMaximum( lround( sb->get_value() + 1 ) );
}

void MainWindow::new_model_()
{
	if( !is_mesh_data_safe_() )
		return;

	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );

	Gtk::Dialog* nd;
	glade_->get_widget( "NewDialog", nd );

	nd->set_transient_for( *mainwindow );

	const int result( nd->run() );

	if( result == Gtk::RESPONSE_OK )
	{
		create_model_();
	}

	nd->hide();
}

void MainWindow::open_model_()
{
	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );

	Gtk::FileChooserDialog fd( "Open Model" );
	fd.set_transient_for( *mainwindow );

	ModelPreview preview( fd );
	fd.set_preview_widget( preview );

	fd.set_action( Gtk::FILE_CHOOSER_ACTION_OPEN );
	fd.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
	fd.add_button( Gtk::Stock::OPEN, Gtk::RESPONSE_OK );

	{
		Gtk::FileFilter filter;
		filter.set_name( "All Models" );
		filter.add_pattern( "*.lwo" );
		filter.add_pattern( "*.obj" );
		fd.add_filter( filter );
	}

	{
		Gtk::FileFilter filter;
		filter.set_name( "Lightwave (*.lwo)" );
		filter.add_pattern( "*.lwo" );
		fd.add_filter( filter );
	}

	{
		Gtk::FileFilter filter;
		filter.set_name( "Wavefront (*.obj)" );
		filter.add_pattern( "*.obj" );
		fd.add_filter( filter );
	}
	
	{
		Gtk::FileFilter filter;
		filter.set_name( "All files" );
		filter.add_pattern( "*" );
		fd.add_filter( filter );
	}

	const std::string no_path_stored( "no-path-stored" );
	const std::string path( Preferences::Instance().Get( "ModelPath", no_path_stored ) );
	if( path != no_path_stored )
		fd.set_filename( path );

	const int result( fd.run() );

	if( result == Gtk::RESPONSE_OK )
	{
		const std::string filename( fd.get_filename() );
		
		MeshHistory::Instance().Clear();
		Create::File f( filename );
		f.Load( MeshHistory::Instance().GetCurrentMesh() );
		while( f.Step() );
		MeshHistory::Instance().GetCurrentMesh().Create( f, progress_ );
		
		editor_.ResetView();
		editor_.Refresh( true, true );

		Preferences::Instance().Set( "ModelPath", filename );
	}
}

void MainWindow::save_model_()
{
	if( MeshHistory::Instance().GetCurrentMesh().FileName().length() )
	{
		MeshHistory::Instance().GetCurrentMesh().Save
			( MeshHistory::Instance().GetCurrentMesh().FileName() );
	}
	else
	{
		save_model_as_();
	}

	on_history_changed_();
}

void MainWindow::save_model_as_()
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	Gtk::Window* mainwindow;
	glade_->get_widget( "MainWindow", mainwindow );

	Gtk::FileFilter filter_lwo;
	filter_lwo.set_name( "Lightwave (*.lwo)" );
	filter_lwo.add_pattern( "*.lwo" );

	Gtk::FileFilter filter_obj;
	filter_obj.set_name( "Wavefront (*.obj)" );
	filter_obj.add_pattern( "*.obj" );

	Gtk::FileFilter filter_vrml;
	filter_vrml.set_name( "VRML (*.wrl)" );
	filter_vrml.add_pattern( "*.wrl" );
	
	SaveDialog fd( "Save Model", *mainwindow, filter_obj );
	
	fd.add_filter( filter_lwo );
	fd.add_filter( filter_obj );
	fd.add_filter( filter_vrml );		
	fd.set_filename( mesh.FileName() );
	fd.set_filter( filter_obj );

	const int result( fd.run() );
	
	if( result == Gtk::RESPONSE_OK )
	{
		mesh.Save( fd.get_filename() );

		if( fd.get_filter()->get_name().find( "lwo" ) != std::string::npos )
		{
			Gtk::MessageDialog md( *mainwindow, "Colors are not saved in the Lightwave format.", false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK );
			md.run();
		}
	}

	on_history_changed_();
}

void MainWindow::on_enable_brush_curve_toggled_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "EnableBrushCurve", tgb );
	Tools::Instance().SetBrushCurveEnabled( tgb->get_active() );
}

void MainWindow::on_radius_changed_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Radius", s );
	Tools::Instance().Brush()->SetRadius( lround( s->get_value() ) );
	editor_.Refresh( false, false );
}
void MainWindow::on_strength_changed_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Strength", s );
	Tools::Instance().Brush()->SetStrength( s->get_value() );
}
void MainWindow::on_intensity_changed_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Intensity", s );
	Tools::Instance().Brush()->SetIntensity( s->get_value() );
}

void MainWindow::update_brush_()
{
	Brush::BrushBase* brush( Tools::Instance().Brush() );
	Gtk::CheckButton* chb, * sx, * sy, * sz;
	Gtk::ColorButton* cb;
	Gtk::ToggleButton* tgb;

	glade_->get_widget( "Layer", tgb );
	brush->SetLayer( tgb->get_active() );
	
	glade_->get_widget( "Dot", tgb );
	brush->SetDot( tgb->get_active() );

	glade_->get_widget( "MaskBrush", tgb );
	brush->SetMask( tgb->get_active() );

	glade_->get_widget( "Colorize", chb );
	brush->SetColorize( chb->get_active() );

	glade_->get_widget( "BrushColor", cb );
	brush->SetPigment( Color( cb->get_color().get_red_p(),
				  cb->get_color().get_green_p(),
				  cb->get_color().get_blue_p() ) );

	glade_->get_widget( "SymmetryX", sx );
	glade_->get_widget( "SymmetryY", sy );
	glade_->get_widget( "SymmetryZ", sz );
	brush->SetSymmetry( Brush::SymmetryData( sx->get_active(),
						 sy->get_active(),
						 sz->get_active() ) );

	update_brush_sensitivities_();

	editor_.Refresh( false, false );
}

void MainWindow::update_brush_controls_()
{
	Brush::BrushBase* brush( Tools::Instance().Brush() );

	update_brush_sensitivities_();

	Gtk::Scale* hs;
	Gtk::CheckButton* chb;
	Gtk::RadioButton* rb;
	Gtk::ToggleButton* tgb;

	glade_->get_widget( "Radius", hs );
	hs->set_value( brush->BasicRadius() );
	glade_->get_widget( "Strength", hs );
	hs->set_value( brush->BasicStrength() );
	glade_->get_widget( "Intensity", hs );
	hs->set_value( brush->BasicIntensity() );

	glade_->get_widget( "SizePressure", tgb );
	tgb->set_active( brush->RadiusPressure() );
	glade_->get_widget( "StrengthPressure", tgb );
	tgb->set_active( brush->StrengthPressure() );
	glade_->get_widget( "IntensityPressure", tgb );
	tgb->set_active( brush->IntensityPressure() );

	glade_->get_widget( "Layer", chb );
	chb->set_active( brush->Layer() );
	glade_->get_widget( "Dot", chb );
	chb->set_active( brush->Dot() );
	glade_->get_widget( "MaskBrush", chb );
	chb->set_active( brush->Mask() );

	glade_->get_widget( "BrushAdd", rb );
	rb->set_active( brush->Direction() == SharpConstruct::Brush::OUT );
	glade_->get_widget( "BrushSub", rb );
	rb->set_active( brush->Direction() == SharpConstruct::Brush::IN );
	glade_->get_widget( "BrushNoDirection", rb );
	rb->set_active( brush->Direction() == SharpConstruct::Brush::OFF );
	glade_->get_widget( "Colorize", chb );
	chb->set_active( brush->Colorize() );

	Gtk::TreeView* tv;
	glade_->get_widget( "BrushShapes", tv );
	tv->get_selection()->select( brush->ShapePath() );
}

void MainWindow::update_brush_direction_()
{
	Gtk::RadioButton* rb;
	glade_->get_widget( "BrushAdd", rb );
	if( rb->get_active() )
		Tools::Instance().Brush()->SetDirection( SharpConstruct::Brush::OUT );
	glade_->get_widget( "BrushSub", rb );
	if( rb->get_active() )
		Tools::Instance().Brush()->SetDirection( SharpConstruct::Brush::IN );
	glade_->get_widget( "BrushNoDirection", rb );
	if( rb->get_active() )
		Tools::Instance().Brush()->SetDirection( SharpConstruct::Brush::OFF );
}

void MainWindow::update_brush_preview_()
{
	Gtk::Image* img;
	glade_->get_widget( "BrushShapePreview", img );

	const int sz( 72 );
	img->set( brush_shape_->scale_simple( sz, sz, Gdk::INTERP_BILINEAR ) );
}

void MainWindow::update_brush_sensitivities_()
{
	const Brush::SupportedProperties& props( Tools::Instance().Brush()->Properties() );

	Gtk::Widget* w;
	glade_->get_widget( "StrengthLabel", w );
	w->set_sensitive( props.Strength );
	glade_->get_widget( "Strength", w );
	w->set_sensitive( props.Strength );
	glade_->get_widget( "StrengthPressure", w );
	w->set_sensitive( props.Strength );
	glade_->get_widget( "IntensityLabel", w );
	w->set_sensitive( props.Intensity );
	glade_->get_widget( "Intensity", w );
	w->set_sensitive( props.Intensity );
	glade_->get_widget( "IntensityPressure", w );
	w->set_sensitive( props.Intensity );
	glade_->get_widget( "Layer", w );
	w->set_sensitive( props.Layer );
	glade_->get_widget( "Dot", w );
	w->set_sensitive( props.Dot );
	glade_->get_widget( "MaskBrush", w );
	w->set_sensitive( props.Mask );
	glade_->get_widget( "BrushAdd", w );
	w->set_sensitive( props.Add );
	glade_->get_widget( "BrushSub", w );
	w->set_sensitive( props.Sub );
	glade_->get_widget( "BrushNoDirection", w );
	w->set_sensitive( props.Add || props.Sub );
	glade_->get_widget( "Colorize", w );
	w->set_sensitive( props.Colorize );
	glade_->get_widget( "BrushColor", w );
	w->set_sensitive( props.Pigment );
}

void MainWindow::update_brush_shape_()
{
	Gtk::TreeView* tv;
	glade_->get_widget( "BrushShapes", tv );
	Gtk::Image* img;
	glade_->get_widget( "BrushShapePreview", img );

	Gtk::ListStore::iterator iter( tv->get_selection()->get_selected() );
	if( iter )
	{
		Gtk::TreeModel::Row row = *iter;
		brush_shape_ = row[ brush_shape_columns_.Shape ];
		Tools::Instance().Brush()->SetShape( brush_shape_, Gtk::TreePath( iter ) );

		update_brush_preview_();
	}
}

void MainWindow::update_graphics_mode_settings_()
{
	Gtk::RadioButton* rb;

	glade_->get_widget( "VBO", rb );
	if( rb->get_active() )
		editor_.DisplayOptions().SetGraphicsMode( Editor::DisplayOptionsData::VBO );
	
	glade_->get_widget( "VAR", rb );
	if( rb->get_active() )
		editor_.DisplayOptions().SetGraphicsMode( Editor::DisplayOptionsData::VAR );
}

void MainWindow::update_lighting_()
{
	Gtk::ColorButton* cb;
	Gtk::Scale* s;

	glade_->get_widget( "DiffuseColor", cb );
	editor_.Lighting().SetDiffuse( cb->get_color() );

	glade_->get_widget( "SpecularColor", cb );
	editor_.Lighting().SetSpecular( cb->get_color() );

	glade_->get_widget( "Shininess", s );
	editor_.Lighting().SetShininess( lroundf( s->get_value() ) );

	editor_.Refresh( false, false );
}

void MainWindow::on_radius_pressure_changed_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "SizePressure", tgb );
	Tools::Instance().Brush()->SetRadiusPressure( tgb->get_active() );
}

void MainWindow::on_strength_pressure_changed_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "StrengthPressure", tgb );
	Tools::Instance().Brush()->SetStrengthPressure( tgb->get_active() );
}

void MainWindow::on_intensity_pressure_changed_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "IntensityPressure", tgb );
	Tools::Instance().Brush()->SetIntensityPressure( tgb->get_active() );
}

void MainWindow::update_progress_bar_( const Progress& p )
{
	/*progress_bar_.set_fraction( p.Amount() );
	progress_bar_.set_text( p.Text() );

	static double PrevTime( Time() );
	if( Time() > PrevTime + 1 )
	{
		Gtk::Window* mw;
		glade_->get_widget( "MainWindow", mw );
		
		while( gtk_events_pending() )
			gtk_main_iteration();

		PrevTime = Time();
		}*/
	
	/*Glib::RefPtr<Gdk::Window> win = mw->get_window();
	{
		Gdk::Rectangle r(0, 0, mw->get_allocation().get_width(),
				 mw->get_allocation().get_height());
		win->invalidate_rect(r, true);
	}
	gdk_window_process_all_updates();*/
}

void MainWindow::update_zoom_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Zoom", s );
	editor_.SetZoom( lround( s->get_value() ) );
}

void MainWindow::zoom_in_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Zoom", s );

	int zoom( lroundf( s->get_value() ) + 1 );
	s->set_value( std::min( zoom, 40 ) );
}

void MainWindow::zoom_out_()
{
	Gtk::Scale* s;
	glade_->get_widget( "Zoom", s );

	int zoom( lroundf( s->get_value() ) - 1 );
	s->set_value( std::max( zoom, 1 ) );
}

void MainWindow::shader_plain_()
{
	editor_.SetShader( "Main" );
}

void MainWindow::shader_normals_()
{
	editor_.SetShader( "Normals" );
}

void MainWindow::on_draw_clicked_()
{
	Tools::Instance().SetDisplaceBrush();
	update_brush_controls_();
}

void MainWindow::on_smooth_clicked_()
{
	Tools::Instance().SetSmoothBrush();
	update_brush_controls_();
}

void MainWindow::on_pinch_clicked_()
{
	Tools::Instance().SetPinchBrush();
	update_brush_controls_();
}

void MainWindow::on_grab_clicked_()
{
	Tools::Instance().SetGrabBrush();
	update_brush_controls_();
}

void MainWindow::apply_interactive_deformation_()
{
	if( Tools::Instance().InDeformationMode() )
		Tools::Instance().Deformation()->Apply();

	Gtk::ToggleButton* tgb;
	glade_->get_widget( "InteractiveOff", tgb );
	tgb->set_active( true );
}

void MainWindow::update_interactive_deformation_()
{
	Gtk::ToggleButton* tgb;
	glade_->get_widget( "InteractiveOff", tgb );
	Gtk::Widget* w;
	glade_->get_widget( "InteractiveApply", w );


	if( tgb->get_active() )
	{
		Tools::Instance().DeactivateDeformation();
		w->set_sensitive( false );
	}
	else
	{
		glade_->get_widget( "Scale", tgb );
		if( tgb->get_active() )
			Tools::Instance().SetDeformation< Scale >();
		else
		{
			glade_->get_widget( "Move", tgb );
			if( tgb->get_active() )
				Tools::Instance().SetDeformation< Move >();
			else
			{
				glade_->get_widget( "Mirror", tgb );
				if( tgb->get_active() )
					Tools::Instance().SetDeformation< Mirror >();
			}
		}
		w->set_sensitive( true );
	}

	editor_.Refresh( false, false );
}
