/* Copyright 2005 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "MeshHistory.hh"
#include "Progress.hh"
#include "Utilities.hh"
#include <algorithm>
#include <limits>

#include <iostream>

using namespace SharpConstruct;

void Mesh::Move( float x, float y, float z )
{	
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
	{
		const float mask( 1.0f - VertexColors()[ i ].Mask() );
		VertexLocations()[ i ].X() += x * mask;
		VertexLocations()[ i ].Y() += y * mask;
		VertexLocations()[ i ].Z() += z * mask;
	}
	RecalculateAllNormals();
	CurrentGrid().SetDivision( 10 );
	ChangeOperation( false, true );

	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

void Mesh::Resize( float x, float y, float z )
{
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
	{
		if( x > 0 )
			VertexLocations()[ i ].X() *= x;
		else if( x < 0 )
			VertexLocations()[ i ].X() /= -x;
		if( y > 0 )
			VertexLocations()[ i ].Y() *= y;
		else if( y < 0 )
			VertexLocations()[ i ].Y() /= -y;
		if( z > 0 )
			VertexLocations()[ i ].Z() *= z;
		else if( z < 0 )
			VertexLocations()[ i ].Z() /= -z;
	}
	
	RecalculateAllNormals();
	Mesh::CurrentGrid().SetDivision( 10 );
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

/** Smooths the entire mesh.
 *
 * Smooths every point in the mesh. Accepts a strength value between zero and
 * one.
**/
void Mesh::Smooth( float strength )
{
	Point3D average;
	
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
	{
		if( VertexUsers()[ i ].Total() > 2 )
		{
			average = VertexNeighborLocAverage( i );
			VertexLocations()[ i ] += ( average - VertexLocations()[ i ] ) * strength * Mask( i );
		}
	}
	
	RecalculateAllNormals();
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

/**
 *
**/
void Mesh::Deflate( float v )
{
	Point3D nloc;
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
	{
		VertexLocations()[ i ] += VertexNormals()[ i ] * v * Mask( i );
	}
	RecalculateAllNormals();
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

/** Normalizes the distances of vertices from the model's center
 *
 */
void Mesh::Spherize( float strength )
{
	UpdateBounds();

	const float sub( 1 - strength );
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
	{
		Point3D& v( VertexLocations()[ i ] );
		Point3D copy( v );
		Normalize( copy );
		v = v * sub + copy * strength * Mask( i );
	}

	RecalculateAllNormals();
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

/** Floods the entire mesh with the current color.
 *
 * All vertex colors are shifted towards the current color. Accepts an intensity
 * value between 0 and 1.
**/
void Mesh::Flood( const Color& color, const float intensity )
{
	for( unsigned i = 0; i < _visible_elements.Vertices; ++i )
		VertexColors()[ i ] += ( color - VertexColors()[ i ] ) * intensity * Mask( i );
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( false, true, false );
}

struct subtri
{
	unsigned O[ 3 ];
	unsigned P[ 3 ];
};
struct subquad
{
	unsigned O[ 4 ];
	unsigned P[ 4 ];
};

void Mesh::Subdivide( bool smooth )
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

	const unsigned tri_size( mesh.Triangles().size() );
	for( unsigned i = 0; i < tri_size; ++i )
	{
		const unsigned poly( i );
		bool unmasked( true );
		for( unsigned j = 0; j <= 2; ++j )
			if( VertexColors()[ Triangles()[ poly ].V[ j ] ].Mask() > 0.5f )
			{
				unmasked = false;
				break;
			}
		
		if( unmasked )
		{
			for( unsigned j = 0; j <= 2; ++j )
			{
				Edge& edge( mesh.Edges().Edges[ Edges().Triangles[ poly ][ j ] ] );
				if( !edge.HasMid )
					add_edge_midpoint_( edge );
			}
		}
	}

	const unsigned quad_size( mesh.Quads().size() );
	for( unsigned i = 0; i < quad_size; ++i )
	{
		const unsigned poly( i );
		bool unmasked( true );
		for( unsigned j = 0; j <= 3; ++j )
			if( VertexColors()[ Quads()[ poly ].V[ j ] ].Mask() > 0.5f )
			{
				unmasked = false;
				break;
			}
		
		if( unmasked )
		{
			for( unsigned j = 0; j <= 3; ++j )
			{
				Edge& edge( mesh.Edges().Edges[ Edges().Quads[ poly ].E[ j ] ] );
				if( !edge.HasMid )
					add_edge_midpoint_( edge );
			}
		}
	}

	generate_subdivided_polygons_();
	mesh.MakeSane();

	//if( smooth )
	//	Smooth( 1 );
	if( smooth )
	{
		Point3DVector nv( VertexLocations() );
		for( unsigned i = 0; i < VertexLocations().size(); ++i )
		{
			unsigned count = 0, j;

			nv[ i ] = Point3D( 0, 0, 0 );
			//nv[ i ] = VertexLocations()[ i ];
			for( j = 0; j < VertexUsers()[ i ].Triangles.size(); j++ )
				nv[ i ] += PolygonCenter( Triangles()[ VertexUsers()[ i ].Triangles[ j ] ] );
			count += j;

			for( j = 0; j < VertexUsers()[ i ].Quads.size(); j++ )
				nv[ i ] += PolygonCenter( Quads()[ VertexUsers()[ i ].Quads[ j ] ] );
			count += j;

			nv[ i ] /= count;
		}
		VertexLocations() = nv;
	}

	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, true, true );
}

void Mesh::AutomaticSubdivide()
{
	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		const float TOO_LONG( 2 );

		float lengths[ 3 ];
		// Find length of each edge
		for( unsigned j = 0; j <= 2; ++j )
		{
			Edge& edge( Edges().Edges[ Edges().Triangles[ i ][ j ] ] );
			lengths[ j ] = Distance( VertexLocations()[ edge.V[ 0 ] ], VertexLocations()[ edge.V[ 1 ] ] );
		}

		unsigned shortest = 0;
		// Find shortest edge
		for( unsigned j = 1; j <= 2; ++j )
		{
			if( lengths[ j ] < lengths[ shortest ] )
				shortest = j;
		}

		// Mark long edges for subdivision
		for( unsigned j = 0; j <= 2; ++j )
		{
			Edge& edge( Edges().Edges[ Edges().Triangles[ i ][ j ] ] );
			if( !edge.HasMid && lengths[ j ] > lengths[ shortest ] * TOO_LONG )
				add_edge_midpoint_( edge );
		}
	}

	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		const float TOO_LONG( 1.5 );

		float lengths[ 4 ];
		// Find length of each edge
		for( unsigned j = 0; j <= 3; ++j )
		{
			Edge& edge( Edges().Edges[ Edges().Quads[ i ].E[ j ] ] );
			lengths[ j ] = Distance( VertexLocations()[ edge.V[ 0 ] ], VertexLocations()[ edge.V[ 1 ] ] );
		}

		unsigned shortest = 0;
		// Find shortest edge
		for( unsigned j = 1; j <= 3; ++j )
		{
			if( lengths[ j ] < lengths[ shortest ] )
				shortest = j;
		}

		// Mark long edges for subdivision
		for( unsigned j = 0; j <= 3; ++j )
		{
			Edge& edge( Edges().Edges[ Edges().Quads[ i ].E[ j ] ] );
			if( !edge.HasMid && lengths[ j ] > lengths[ shortest ] * TOO_LONG )
				add_edge_midpoint_( edge );
		}
	}

	generate_subdivided_polygons_();
	MakeSane();
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, true, true );
}

void Mesh::generate_subdivided_polygons_()
{
	for( unsigned i = 0; i < Edges().Triangles.size(); ++i )
	{
		const Triangle poly( Triangles()[ i ] );
		const TriangleEdges& es( Edges().Triangles[ i ] );
		const unsigned count( SubCount( es, Edges().Edges ) );

		if( count == 3 )
		{
			/*
			     /\
			    /__\
			   /\  /\
			  /__\/__\

			*/

			Triangles().push_back( Triangle( poly.V[ 0 ], Edges().Edges[ es[ 0 ] ].Mid, Edges().Edges[ es[ 2 ] ].Mid ) );
			Triangles().push_back( Triangle( poly.V[ 1 ], Edges().Edges[ es[ 1 ] ].Mid, Edges().Edges[ es[ 0 ] ].Mid ) );
			Triangles().push_back( Triangle( poly.V[ 2 ], Edges().Edges[ es[ 2 ] ].Mid, Edges().Edges[ es[ 1 ] ].Mid ) );
			Triangles()[ i ] =     Triangle( Edges().Edges[ es[ 0 ] ].Mid, Edges().Edges[ es[ 1 ] ].Mid, Edges().Edges[ es[ 2 ] ].Mid );
		}
		else if( count == 2 )
		{
			/*    .
			     / \
			    /___\
			   /     \
			  /_______\

			*/

			unsigned opp( 0 );
			for( unsigned j = 0; j < 3; ++j )
				if( !Edges().Edges[ es[ j ] ].HasMid )
					opp = j;

			unsigned hit1( opp + 1 );
			if( hit1 > 2 )
				hit1 -= 3;

			unsigned hit2( hit1 + 1 );
			if( hit2 > 2 )
				hit2 -= 3;

			Quads().push_back( Quad( poly.V[ opp ], poly.V[ hit1 ], Edges().Edges[ es[ hit1 ] ].Mid, Edges().Edges[ es[ hit2 ] ].Mid ) );
			Triangles()[ i ] = Triangle( poly.V[ hit2 ], Edges().Edges[ es[ hit2 ] ].Mid, Edges().Edges[ es[ hit1 ] ].Mid );
		}
		else if( count == 1 )
		{	
			/*   .
			    /|\
			   / | \
			  /__|__\

			*/

			unsigned hit( 0 );
			for( unsigned j = 0; j < 3; ++j )
				if( Edges().Edges[ es[ j ] ].HasMid )
					hit = j;

			unsigned opp( hit + 2 );
			if( opp > 2 )
				opp -= 3;

			unsigned n1( hit + 1 );
			if( n1 > 2 )
				n1 -= 3;

			Triangles().push_back( Triangle( Edges().Edges[ es[ hit ] ].Mid, poly.V[ n1 ], poly.V[ opp ] ) );
			Triangles()[ i ] =     Triangle( Edges().Edges[ es[ hit ] ].Mid, poly.V[ opp ], poly.V[ hit ] );
		}
	}

	for( unsigned i = 0; i < Edges().Quads.size(); ++i )
	{
		const Quad poly( Quads()[ i ] );
		const QuadEdges& es( Edges().Quads[ i ] );
		const unsigned count( SubCount( es, Edges().Edges ) );
		/* There are four different quad subdivision cases, based on which
		   Edges().Edges need to be cut. */
		if( count == 4 )
		{
			// Create center
			const unsigned cntr_ndx( VertexLocations().size() );
			VertexLocations().push_back( PolygonCenter( poly ) );
			VertexColors().push_back( MixColors( VertexColors()[ poly.V[ 0 ] ],
							     VertexColors()[ poly.V[ 1 ] ],
							     VertexColors()[ poly.V[ 2 ] ],
							     VertexColors()[ poly.V[ 3 ] ] ) );

			Quads().push_back( Quad( cntr_ndx, Edges().Edges[ es.E[ 3 ] ].Mid, poly.V[ 0 ], Edges().Edges[ es.E[ 0 ] ].Mid ) );
			Quads().push_back( Quad( cntr_ndx, Edges().Edges[ es.E[ 0 ] ].Mid, poly.V[ 1 ], Edges().Edges[ es.E[ 1 ] ].Mid ) );
			Quads().push_back( Quad( cntr_ndx, Edges().Edges[ es.E[ 1 ] ].Mid, poly.V[ 2 ], Edges().Edges[ es.E[ 2 ] ].Mid ) );
			Quads()[ i ] =     Quad( cntr_ndx, Edges().Edges[ es.E[ 2 ] ].Mid, poly.V[ 3 ], Edges().Edges[ es.E[ 3 ] ].Mid );
		}
		else if( count == 3 )
		{
			unsigned unhit( 0 );
			// Find the side that isn't subd
			for( unsigned j = 0; j < 4; ++j )
				if( !Edges().Edges[ es.E[ j ] ].HasMid )
					unhit = j;

			unsigned hit1( unhit + 1 );
			if( hit1 > 3 )
				hit1 -= 4;
			unsigned hit2( hit1 + 1 );
			if( hit2 > 3 )
				hit2 -= 4;
			unsigned hit3( hit2 + 1 );
			if( hit3 > 3 )
				hit3 -= 4;

			Triangles().push_back( Triangle( Edges().Edges[ es.E[ hit1 ] ].Mid, Edges().Edges[ es.E[ hit2 ] ].Mid, Edges().Edges[ es.E[ hit3 ] ].Mid ) );
			Triangles().push_back( Triangle( Edges().Edges[ es.E[ hit1 ] ].Mid, poly.V[ hit2 ], Edges().Edges[ es.E[ hit2 ] ].Mid ) );
			Triangles().push_back( Triangle( poly.V[ hit3 ], Edges().Edges[ es.E[ hit3 ] ].Mid, Edges().Edges[ es.E[ hit2 ] ].Mid ) );
			Quads()[ i ] = Quad( poly.V[ unhit ], poly.V[ hit1 ], Edges().Edges[ es.E[ hit1 ] ].Mid, Edges().Edges[ es.E[ hit3 ] ].Mid );
		}
		else if( count == 2 )
		{
			unsigned first( 0 );
			for( unsigned j = 0; j < 4; ++j )
				if( Edges().Edges[ es.E[ j ] ].HasMid )
					first = j;

			unsigned check( first + 2 );
			if( check > 3 )
				check -= 4;

			if( !Edges().Edges[ es.E[ check ] ].HasMid ) // Corner is being subdivided
			{
				unsigned opp( 0 );
				for( unsigned j = 0; j < 4; ++j )
				{
					const unsigned j2( j == 3 ? 0 : j + 1 );
					if( !Edges().Edges[ es.E[ j ] ].HasMid &&
					    !Edges().Edges[ es.E[ j2 ] ].HasMid )
						opp = j2;
				}
				unsigned hit1( opp + 1 );
				if( hit1 > 3 )
					hit1 -= 4;
				unsigned hit2( opp + 2 );
				if( hit2 > 3 )
					hit2 -= 4;
				unsigned n1( opp + 1 );
				if( n1 > 3 )
					n1 -= 4;
				unsigned n2( hit2 + 1 );
				if( n2 > 3 )
					n2 -= 4;
				
				Triangles().push_back( Triangle( Edges().Edges[ es.E[ hit2 ] ].Mid, poly.V[ n2 ], poly.V[ opp ] ) );
				Triangles().push_back( Triangle( poly.V[ n1 ], Edges().Edges[ es.E[ hit1 ] ].Mid, poly.V[ opp ] ) );
				Quads()[ i ] = Quad( poly.V[ opp ], Edges().Edges[ es.E[ hit1 ] ].Mid, poly.V[ hit2 ], Edges().Edges[ es.E[ hit2 ] ].Mid );
			}
			else
			{
				unsigned f2( first + 1 );
				if( f2 > 3 )
					f2 -= 4;
				unsigned sec( first + 2 );
				if( sec > 3 )
					sec -= 4;
				unsigned s2( sec + 1 );
				if( s2 > 3 )
					s2 -= 4;
				Quads().push_back( Quad( Edges().Edges[ es.E[ first ] ].Mid, poly.V[ f2 ], poly.V[ sec ], Edges().Edges[ es.E[ sec ] ].Mid ) );
				Quads()[ i ] =     Quad( Edges().Edges[ es.E[ sec ] ].Mid, poly.V[ s2 ], poly.V[ first ], Edges().Edges[ es.E[ first ] ].Mid );
			}
		}
		else if( count == 1 )
		{
			// Find the side that is subd
			unsigned hit( 0 );
			for( unsigned j = 0; j < 4; ++j )
				if( Edges().Edges[ es.E[ j ] ].HasMid )
					hit = j;

			// Find the point to which we will connect
			unsigned con( hit + 3 );
			if( con > 3 )
				con -= 4;

			Triangles().push_back( Triangle( poly.V[ hit ], Edges().Edges[ es.E[ hit ] ].Mid, poly.V[ con ] ) );
			Quads()[ i ].V[ hit ] = Edges().Edges[ es.E[ hit ] ].Mid;
		}
	}
}

#if 0
void Mesh::Subdivide( bool smooth )
{
	std::vector< subtri > st( Triangles().size() );
	unsigned lndx = VertexLocations().size();

	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		// Copy Triangles() data to st
		for( unsigned j = 0; j <= 2; j++ )
			st[ i ].O[ j ] = Triangles()[ i ].V[ j ];

		// For each edge
		for( unsigned j = 0; j <= 2; j++ )
		{
			int j2 = j + 1;
			if( j2 > 2 )
				j2 = 0;
			
			// Find all polygon users
			std::map< unsigned, unsigned > tusrs = _triangle_users( st[ i ].O[ j ], st[ i ].O[ j2 ] );
			
			std::map< unsigned, unsigned >::iterator it;
			int subndx = -1;
			
			for( std::map< unsigned, unsigned >::iterator it = tusrs.begin();
			     it != tusrs.end(); it++ )
			{
				if( ( *it ).first < i )
				{
					it = tusrs.begin();
					subndx = st[ ( *it ).first ].P[ ( *it ).second - 1 ];
					break;
				}
			}
			
			// If not found, calculate a new midpoint
			if( subndx == -1 )
			{
				VertexLocations().resize( VertexLocations().size() + 1 );
				VertexLocations()[ lndx ] = Midpoint( VertexLocations()[ st[ i ].O[ j ] ],
													VertexLocations()[ st[ i ].O[ j2 ] ] );
				VertexColors().resize( VertexColors().size() + 1 );
				VertexColors()[ lndx ] = MixColors( VertexColors()[ st[ i ].O[ j ] ],
				                                    VertexColors()[ st[ i ].O[ j2 ] ] );
				st[ i ].P[ j ] = lndx++;
			}
			else
			{
				st[ i ].P[ j ] = subndx;
			}
		}
	}
	
	std::vector< subquad > sq( Quads().size() );
	lndx = VertexLocations().size();
	
	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		// Copy quad data to sq
		for( unsigned j = 0; j <= 3; j++ )
			sq[ i ].O[ j ] = Quads()[ i ].V[ j ];
		
		// For each edge
		for( unsigned j = 0; j <= 3; j++ )
		{
			int j2 = j + 1;
			if( j2 > 3 )
				j2 = 0;
			
			// Find all polygon users
			std::map< unsigned, unsigned > tusrs = _triangle_users( sq[ i ].O[ j ], sq[ i ].O[ j2 ] );
			
			std::map< unsigned, unsigned >::iterator it;
			int subndx = -1;
			
			if( tusrs.size() )
			{
				it = tusrs.begin();
				subndx = st[ ( *it ).first ].P[ ( *it ).second - 1 ];
			}
			
			// No connected triangles
			if( subndx == -1 )
			{
				std::map< unsigned, unsigned > qusrs = _quad_users( sq[ i ].O[ j ], sq[ i ].O[ j2 ] );
				for( it = qusrs.begin(); it != qusrs.end(); it++ )
				{
					if( ( *it ).first < i )
					{
						subndx = sq[ ( *it ).first ].P[ (int)(( *it ).second) ];
						break;
					}
				}
			}
			
			// If not found, calculate a new midpoint
			if( subndx == -1 )
			{
				VertexLocations().resize( VertexLocations().size() + 1 );
				VertexLocations()[ lndx ] = Midpoint( VertexLocations()[ sq[ i ].O[ j ] ],
								    VertexLocations()[ sq[ i ].O[ j2 ] ] );
				VertexColors().resize( VertexColors().size() + 1 );
				VertexColors()[ lndx ] = MixColors( VertexColors()[ sq[ i ].O[ j ] ],
				                                    VertexColors()[ sq[ i ].O[ j2 ] ] );
				sq[ i ].P[ j ] = lndx++;
			}
			else
			{
				sq[ i ].P[ j ] = subndx;
			}
		}
	}

	Triangles().clear();
	Quads().clear();
	/* This loop makes the following subdivisional pattern:
	 *
	 *    /\               /\
	 *   /  \    ====\    /__\
	 *  /    \   ====/   /\  /\
	 * /______\         /__\/__\
	 *
	 * Currently disabled in favor of a quad-producing algorithm.
	 */
	/*for( unsigned i = 0; i < st.size(); ++i )
	{
		Triangles().push_back( Triangle( st[ i ].O[ 0 ], st[ i ].P[ 0 ], st[ i ].P[ 2 ] ) );
		Triangles().push_back( Triangle( st[ i ].P[ 0 ], st[ i ].O[ 1 ], st[ i ].P[ 1 ] ) );
		Triangles().push_back( Triangle( st[ i ].P[ 1 ], st[ i ].O[ 2 ], st[ i ].P[ 2 ] ) );
		Triangles().push_back( Triangle( st[ i ].P[ 0 ], st[ i ].P[ 1 ], st[ i ].P[ 2 ] ) );
	}*/

	// This loop produces three quads instead of four triangles.
	for( unsigned i = 0; i < st.size(); ++i )
	{
		// Find center
		Point3D cen( 0, 0, 0 );
		for( unsigned j = 0; j <= 2; j++ )
			cen += VertexLocations()[ st[ i ].O[ j ] ];
		cen /= 3;
		
		unsigned cndx = VertexLocations().size();
		VertexLocations().push_back( cen );
		VertexColors().push_back( MixColors( VertexColors()[ st[ i ].O[ 0 ] ],
		                                     VertexColors()[ st[ i ].O[ 1 ] ],
		                                     VertexColors()[ st[ i ].O[ 2 ] ] ) );

		Quads().push_back( Quad( cndx, st[ i ].P[ 2 ], st[ i ].O[ 0 ], st[ i ].P[ 0 ] ) );
		Quads().push_back( Quad( cndx, st[ i ].P[ 0 ], st[ i ].O[ 1 ], st[ i ].P[ 1 ] ) );
		Quads().push_back( Quad( cndx, st[ i ].P[ 1 ], st[ i ].O[ 2 ], st[ i ].P[ 2 ] ) );
	}

	for( unsigned i = 0; i < sq.size(); ++i )
	{
		// Find center
		Point3D cen( 0, 0, 0 );
		for( unsigned j = 0; j <= 3; j++ )
			cen += VertexLocations()[ sq[ i ].O[ j ] ];
		cen /= 4;
		
		unsigned cndx = VertexLocations().size();
		VertexLocations().push_back( cen );
		VertexColors().push_back( MixColors( VertexColors()[ sq[ i ].O[ 0 ] ],
		                                     VertexColors()[ sq[ i ].O[ 1 ] ],
		                                     VertexColors()[ sq[ i ].O[ 2 ] ],
		                                     VertexColors()[ sq[ i ].O[ 3 ] ] ) );
		
		Quads().push_back( Quad( sq[ i ].O[ 0 ], sq[ i ].P[ 0 ], cndx, sq[ i ].P[ 3 ] ) );
		Quads().push_back( Quad( sq[ i ].P[ 0 ], sq[ i ].O[ 1 ], sq[ i ].P[ 1 ], cndx ) );
		Quads().push_back( Quad( sq[ i ].P[ 1 ], sq[ i ].O[ 2 ], sq[ i ].P[ 2 ], cndx ) );
		Quads().push_back( Quad( sq[ i ].P[ 2 ], sq[ i ].O[ 3 ], sq[ i ].P[ 3 ], cndx ) );
	}

	// Return mesh data to a consistent state
	
	UpdateVertexUsers();
	VertexNormals().resize( VertexLocations().size() );
	TriangleNormals().resize( Triangles().size() );
	QuadNormals().resize( Quads().size() );
	//_recalculate_all_normals();
	if( smooth )
	{
		Point3DVector nv( VertexLocations() );
		for( unsigned i = 0; i < VertexLocations().size(); ++i )
		{
			unsigned count = 0, j;

			nv[ i ] = Point3D( 0, 0, 0 );
			//nv[ i ] = VertexLocations()[ i ];
			for( j = 0; j < VertexUsers()[ i ].Triangles.size(); j++ )
				nv[ i ] += PolygonCenter( Triangles()[ VertexUsers()[ i ].Triangles[ j ] ] );
			count += j;

			for( j = 0; j < VertexUsers()[ i ].Quads.size(); j++ )
				nv[ i ] += PolygonCenter( Quads()[ VertexUsers()[ i ].Quads[ j ] ] );
			count += j;

			nv[ i ] /= count;
		}
		VertexLocations() = nv;
	}

	ResetVisibles();

	_visible_elements.RestoreOrder.Enabled = false;


	Mesh::CurrentGrid().SetDivision( 10 );
	
	ChangeOperation( true, true );

	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, true, true );
}
#endif
/*void Mesh::Optimize( float amt )
{
	const int orig = _polygon_normals.size();
	std::vector< std::map< unsigned, float > > dists(
		_vertex_polygon_indices.size() );
	
	// Dists has one map for every vertex; the map contains every other vertex
	// the vertex is connected to (key) and the distance between them as the
	// value
	for( unsigned i = 0; i < VertexLocations().size(); ++i )
	{
		std::set< unsigned > cs = _connected_vertices( i );
		for( std::set< unsigned >::iterator it = cs.begin(); it != cs.end();
			 ++it )
		{
			if( *it < i )
				continue;
			dists[ i ][ *it ] = VertexLocations()[ i ].Distance(
				VertexLocations()[ *it ] );
		}
	}
	while( (int)_indices.size() / 3 > orig - amt )
	{
		unsigned P[ 0 ] = 0, P[ 1 ] = *_connected_vertices( 0 ).begin();
		float shortest = VertexLocations()[ P[ 0 ] ].Distance( VertexLocations()[ P[ 1 ] ] );
		
		for( unsigned i = 1; i < dists.size(); ++i )
		{
			for( std::map< unsigned, float >::iterator it = dists[ i ].begin();
			     it != dists[ i ].end(); ++it )
			{
				if( ( *it ).second < shortest ||
					shortest < std::numeric_limits< float >::epsilon() )
				{
					shortest = ( *it ).second;
					P[ 0 ] = i;
					P[ 1 ] = ( *it ).first;
				}
			}
		}
	
		std::map< unsigned, unsigned > cn = _polygons_with_vertices( P[ 0 ], P[ 1 ] );
		
		// Polygons that use P[ 1 ] should use P[ 0 ] instead
		for( unsigned i = 0; i < _vertex_polygon_indices[ P[ 1 ] ].size(); ++i )
		{
			for( unsigned j = 0; j <= 2; j++ )
			{
				const unsigned p = _vertex_polygon_indices[ P[ 1 ] ][ i ];
				if( _indices[ p * 3 + j ] == P[ 1 ] )
					_indices[ p * 3 + j ] = P[ 0 ];
			}
		}
		
		// Remove polygons that use both P[ 0 ] and P[ 1 ]
		for( std::map< unsigned, unsigned >::reverse_iterator i = cn.rbegin();
			 i != cn.rend(); ++i )
		{
			for( unsigned j = 0; j <= 2; j++ )
			{
				const int ndx = _indices[ ( *i ).first * 3 + j ];
				for( unsigned k = 0; k < _vertex_polygon_indices[ ndx ].size(); k++ )
				{
				*//*	if( _vertex_polygon_indices[ ndx ][ k ] == ( *i ).first )
						_vertex_polygon_indices[ ndx ].erase(
							_vertex_polygon_indices[ ndx ].begin() + k );*//*
				}
			}
			//std::cout << "rm: " << ( *i ).first << std::endl;
			_indices.erase( _indices.begin() + ( *i ).first * 3,
							_indices.begin() + ( *i ).first * 3 + 3 );
		}
		
		// P[ 0 ] = the midpoint of P[ 0 ] and P[ 1 ]
		VertexLocations()[ P[ 0 ] ] = Midpoint( VertexLocations()[ P[ 0 ] ], VertexLocations()[ P[ 1 ] ] );
		
		_vertex_polygon_indices[ P[ 1 ] ].clear();
		_recalculate_vertex_polygon_indices();
		
		dists[ P[ 0 ] ].clear();
		dists[ P[ 1 ] ].clear();
		
		//std::cout << P[ 0 ] << " " << P[ 1 ] << std::endl;
	}
	
	_polygon_normals.resize( _indices.size() / 3 );
	_recalculate_vertex_polygon_indices();
	RecalculateAllNormals();
	
	ChangeOperation();
}*/

int Mesh::RemoveDuplicates()
{
	return 0;
	/*unsigned int i, j, k, l, total = 0;
	
	for( i = 0; i < _vertices.size() - 1; ++i )
		for( j = i + 1; j < _vertices.size(); j++ )
			if( _vertices[ i ].Location == _vertices[ j ].Location )
			{
				// We've found a duplicate, destroy it
				_vertices.erase( _vertices.begin() + j );
				
				for( k = 0; k < _vertex_polygon_indices[ j ].size(); k++ )
				{
					// Copy all of the jth's polygons to i's vector
					_vertex_polygon_indices[ i ].push_back( _vertex_polygon_indices[ j ][ k ] );
					for( l = 0; l < 3; l++ )
						if( _indices[ _vertex_polygon_indices[ j ][ k * 3 + l ] ] == j )
							// Set any index equaling j to i
							_indices[ _vertex_polygon_indices[ j ][ k * 3 + l ] ] == i;
				}
				
				// Destroy the extra vertex_polygon_index
				_vertex_polygon_indices.erase( _vertex_polygon_indices.begin() + j );
				
				for( k = 0; k < _indices.size(); k++ )
					if( _indices[ k ] > j )
						_indices[ k ]--;
				total++;
				j--;
			}
	
	_recalculate_all_normals();
			
	return total;*/
}

void Mesh::Mirror( bool x, bool y, bool z )
{
	if( !x && !y && !z )
		return;
	
	for( unsigned i = 0; i < VertexLocations().size(); ++i )
	{
		if( x )
			VertexLocations()[ i ].X() = -VertexLocations()[ i ].X();
		if( y )
			VertexLocations()[ i ].Y() = -VertexLocations()[ i ].Y();
		if( z )
			VertexLocations()[ i ].Z() = -VertexLocations()[ i ].Z();
	}
	
	if( !( ( x && y && !z ) || ( x && z && !y ) || ( y && z && !x ) ) )
		FlipNormals();
	
	
	RecalculateAllNormals();
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

void Mesh::FlipNormals()
{
	for( unsigned i = 0; i < Triangles().size(); ++i )
		Triangles()[ i ].ReverseVertices();
	for( unsigned i = 0; i < Quads().size(); ++i )
		Quads()[ i ].ReverseVertices();
	
	RecalculateAllNormals();
	ChangeOperation( true, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, true, true );
}

void Mesh::Unify()
{
	UpdateBounds();

	for( unsigned i = 0; i < VertexLocations().size(); ++i )
	{
		VertexLocations()[ i ] -= bounds_.Center();
		VertexLocations()[ i ] /= bounds_.Diameter() / 2;
	}
	ChangeOperation( false, true );
	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, false, false );
}

void Mesh::ToTriangles()
{
	const std::string total( ToString( static_cast< float >
					   ( Quads().size() ) ) );
	std::vector< Quad > nquads;
	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		const Quad quad( Quads()[ i ] );

		// Is quad masked?
		bool masked( false );
		for( unsigned j = 0; j <= 3; ++j )
		{
			if( Mask( quad.V[ j ] ) < 0.5 )
			{
				masked = true;
				break;
			}
		}

		if( masked )
			nquads.push_back( quad );
		else
		{
			Triangles().push_back( Triangle( quad.V[ 0 ],
							 quad.V[ 1 ],
							 quad.V[ 2 ] ) );
			Triangles().push_back( Triangle( quad.V[ 0 ],
							 quad.V[ 2 ],
							 quad.V[ 3 ] ) );
		}
	}
	
	Quads().swap( nquads );

	QuadNormals().resize( Quads().size() );

	ResetVisibles();
	
	TriangleNormals().resize( Triangles().size() );
	UpdateVertexUsers();
	RecalculateAllNormals();

	_visible_elements.RestoreOrder.Enabled = false;
	
	Mesh::CurrentGrid().SetDivision( 10 );
	
	ChangeOperation( true, true );

	MeshHistory::Instance().SignalEntireSectionChanged().emit( true, true, true );
}
