/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "DataDir.hh"
#include "Log.hh"
#include "Shaders.hh"
#include "config.h"

#include <gdkmm/gl/query.h>
#include <fstream>

using SharpConstruct::VertexShader;

VertexShader::VertexShader()
: available_( false )
, shader_name_( "Main" )
{}

void VertexShader::Initialize()
{
	check_extensions_();

	if( Available() )
	{
		using Gdk::GL::get_proc_address;
		glCreateProgramObjectARB  = ( PFNGLCREATEPROGRAMOBJECTARBPROC )
			get_proc_address( "glCreateProgramObjectARB" );
		glCreateShaderObjectARB   = ( PFNGLCREATESHADEROBJECTARBPROC )
			get_proc_address( "glCreateShaderObjectARB");
		glShaderSourceARB         = ( PFNGLSHADERSOURCEARBPROC )
			get_proc_address( "glShaderSourceARB");
		glCompileShaderARB        = ( PFNGLCOMPILESHADERARBPROC )
			get_proc_address( "glCompileShaderARB");
		glAttachObjectARB         = ( PFNGLATTACHOBJECTARBPROC )
			get_proc_address( "glAttachObjectARB");
		glLinkProgramARB          = ( PFNGLLINKPROGRAMARBPROC )
			get_proc_address( "glLinkProgramARB");
		glUseProgramObjectARB     = ( PFNGLUSEPROGRAMOBJECTARBPROC )
			get_proc_address( "glUseProgramObjectARB");
                glDeleteObjectARB         = ( PFNGLDELETEOBJECTARBPROC )
			get_proc_address( "glDeleteObjectARB");
                glDetachObjectARB         = ( PFNGLDETACHOBJECTARBPROC )
			get_proc_address( "glDetachObjectARB");
		glGetInfoLogARB           = ( PFNGLGETINFOLOGARBPROC )
			get_proc_address( "glGetInfoLogARB" );
		glGetObjectParameterivARB = ( PFNGLGETOBJECTPARAMETERIVARBPROC )
			get_proc_address( "glGetObjectParameterivARB" );
	}
}

bool VertexShader::Available() const
{
	return available_;
}

std::string VertexShader::Log() const
{
	if( Available() )
	{
		int length( 0 );
		glGetObjectParameterivARB( vertex_shader_, GL_OBJECT_INFO_LOG_LENGTH_ARB,
					   &length );

		std::string ret;
		if( length > 0 )
		{
			std::vector< char > log( length );
			glGetInfoLogARB( vertex_shader_, log.size(), 0, &log[ 0 ] );
			for( unsigned i = 0; i < log.size(); ++i )
				ret += log[ i ];
		}

		if( ret.length() < 2 )
			return "";
		else
			return ret;
	}
	else
		return "";
}

void VertexShader::Create()
{
	if( Available() )
	{
		vertex_shader_ = glCreateShaderObjectARB( GL_VERTEX_SHADER_ARB );

		const int length( -1 );
                std::string shaderfilename( DataDir( "/sharpconstruct/shaders/" ) );
                shaderfilename += shader_name_;
                shaderfilename += ".glsl";
		const std::string file( load_shader_( shaderfilename ) );
		const char* cptr( file.c_str() );
		glShaderSourceARB( vertex_shader_, 1, &cptr, &length );
		glCompileShaderARB( vertex_shader_ );
		
		program_ = glCreateProgramObjectARB();
		glAttachObjectARB( program_, vertex_shader_ );
		
		glLinkProgramARB( program_ );

                std::string log(Log());
                if(log.length())
			Log::Instance().Add( "Shader error: " + log );
	        else
		        Log::Instance().Add( "Vertex shader created successfully" );

		glEnable( GL_VERTEX_PROGRAM_TWO_SIDE_ARB );
	}
};

void VertexShader::Destroy()
{
	if( Available() )
        {
		glDetachObjectARB( program_, vertex_shader_ );
		glDeleteObjectARB( vertex_shader_ );
		glDeleteObjectARB( program_ );
        }
}

void VertexShader::Bind()
{
	if( Available() )
		glUseProgramObjectARB( program_ );
}

void VertexShader::Unbind()
{
	if( Available() )
		glUseProgramObjectARB( 0 );
}

void VertexShader::SetShaderName(std::string shadername)
{
	if( Available() )
        {
          Destroy();
          shader_name_ = shadername;
          Create();
        }
}

void VertexShader::check_extensions_()
{
	available_ = Gdk::GL::query_gl_extension( "GL_ARB_shader_objects" );
}

const std::string VertexShader::load_shader_( const std::string& filename )
{
	std::string shader = Glib::file_get_contents( filename );
	return shader;
}
